"""Class implementation for an array.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.boolean import Boolean
from apysc._type.copy_interface import CopyInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface

T = TypeVar('T')


class Array(
        CopyInterface, RevertInterface, CustomEventInterface, Generic[T]):
    """
    Array class for the apysc library.

    References
    ----------
    - Array document
        - https://simon-ritchie.github.io/apysc/array.html
    - Array class comparison interfaces document
        - https://simon-ritchie.github.io/apysc/array_comparison.html

    Examples
    --------
    >>> import apysc as ap
    >>> arr: ap.Array = ap.Array([1, 2, 3])
    >>> arr
    Array([1, 2, 3])

    >>> arr[0]
    1

    >>> arr[1]
    2

    >>> arr = ap.Array((4, 5, 6))
    >>> arr
    Array([4, 5, 6])

    >>> arr = ap.Array(range(3))
    >>> arr
    Array([0, 1, 2])

    >>> arr = ap.Array([1, 2, 3])
    >>> arr.append(4)
    >>> arr
    Array([1, 2, 3, 4])

    >>> arr = ap.Array([1, 2, 3])
    >>> arr = arr.concat([4, 5, 6])
    >>> arr
    Array([1, 2, 3, 4, 5, 6])
    """

    _initial_value: Union[List[Any], tuple, 'Array']
    _value: List[Any]

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __init__(
            self,
            value: Union[List[T], tuple, range, 'Array']) -> None:
        """
        Array class for the apysc library.

        Parameters
        ----------
        value : Array or list or tuple or range
            Initial array value.

        References
        ----------
        - Array document
            - https://simon-ritchie.github.io/apysc/array.html
        - Array class comparison interfaces document
            - https://simon-ritchie.github.io/apysc/array_comparison.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr
        Array([1, 2, 3])

        >>> arr[0]
        1

        >>> arr[1]
        2

        >>> arr = ap.Array((4, 5, 6))
        >>> arr
        Array([4, 5, 6])

        >>> arr = ap.Array(range(3))
        >>> arr
        Array([0, 1, 2])
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._expression.event_handler_scope import \
            TemporaryNotHandlerScope
        with TemporaryNotHandlerScope():
            TYPE_NAME: str = var_names.ARRAY
            self._validate_acceptable_value_type(value=value)
            value = self._convert_range_to_list(value=value)
            value_: Union[List[Any], tuple, 'Array'] = value
            self._initial_value = value_
            self._type_name = TYPE_NAME
            self._value = self._get_list_value(value=value)
            self.variable_name = expression_variables_util.\
                get_next_variable_name(type_name=TYPE_NAME)
            self._append_constructor_expression()

    def _convert_range_to_list(
            self, *,
            value: Any) -> Union[List[Any], tuple, 'Array']:
        """
        Convert argument value to list if a specified
        value is a range type.

        Parameters
        ----------
        value : Array or list or tuple or range
            Target value.

        Returns
        -------
        value : Array or list or tuple
            Converted value.
        """
        if isinstance(value, range):
            return list(value)
        return value

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        from apysc._type import value_util
        expression: str = f'var {self.variable_name} = '
        if isinstance(self._initial_value, Array):
            expression += f'{self._initial_value.variable_name};'
        else:
            value_str: str = value_util.get_value_str_for_expression(
                value=self._value)
            expression += f'{value_str};'
        ap.append_js_expression(expression=expression)

    def _get_list_value(
            self, *, value: Union[List[Any], tuple, 'Array']) -> List[Any]:
        """
        Get a list value from a specified list, tuple, or
        Array value.

        Parameters
        ----------
        value : Array or list or tuple
            Specified list, tuple, or Array value.

        Returns
        -------
        list_val : list
            Converted list value.
        """
        if isinstance(value, tuple):
            return list(value)
        if isinstance(value, Array):
            return value._value
        return list(value)  # type: ignore[arg-type]

    def _validate_acceptable_value_type(
            self, *,
            value: Union[List[Any], tuple, range, 'Array']) -> None:
        """
        Validate whether a specified value is an acceptable type
        or not.

        Parameters
        ----------
        value : Array or list or tuple or range
            Iterable value to check.

        Raises
        ------
        ValueError
            If specified value's type is not list, tuple,
            or Array value.
        """
        if isinstance(value, (list, tuple, range, Array)):
            return
        raise ValueError(
            'Not acceptable value type is specified.'
            f'\nSpecified value type: {type(value)}'
            '\nAcceptable types: list, tuple, range, and Array')

    @property
    def value(self) -> Union[List[Any], tuple, 'Array']:
        """
        Get a current array value.

        Returns
        -------
        value : list
            Current array value.

        References
        ----------
        - apysc fundamental data classes value interface
            - https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.value = [4, 5, 6]
        >>> arr.value
        [4, 5, 6]
        """
        return self._value

    @value.setter
    def value(self, value: Union[List[Any], tuple, 'Array']) -> None:
        """
        Set array value.

        Parameters
        ----------
        value : Array or list or tuple
            Iterable value (list, tuple, or Array) to set.

        References
        ----------
        apysc fundamental data classes value interface
            https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html  # noqa
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='value', args=[value], kwargs={},
                module_name=__name__,
                class_name=Array.__name__):
            self._validate_acceptable_value_type(value=value)
            self._value = self._get_list_value(value=value)
            self._append_value_setter_expression(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_value_setter_expression(
            self, *, value: Union[List[Any], tuple, 'Array']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : Array or list or tuple
            Iterable value (Array, list, or tuple) to set.
        """
        import apysc as ap
        from apysc._type import value_util
        expression: str = f'{self.variable_name} = '
        if isinstance(value, Array):
            expression += f'{value.variable_name};'
        else:
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression += f'{value_str};'
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def append(self, value: T) -> None:
        """
        Add any value to the end of this array.
        This method behaves the same `push` method.

        Parameters
        ----------
        value : *
            Any value to append.

        References
        ----------
        - Array class append and push interfaces document
            - https://simon-ritchie.github.io/apysc/array_append_and_push.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.append(4)
        >>> arr
        Array([1, 2, 3, 4])
        """
        self._value.append(value)
        self._append_push_and_append_expression(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def push(self, value: T) -> None:
        """
        Add any value to the end of this array. This
        interface behaves the same as the `append` method.

        Parameters
        ----------
        value : *
            Any value to append.

        References
        ----------
        - Array class append and push interfaces document
            - https://simon-ritchie.github.io/apysc/array_append_and_push.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.push(4)
        >>> arr
        Array([1, 2, 3, 4])
        """
        self.append(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_push_and_append_expression(
            self, *, value: T) -> None:
        """
        Append push and append method expression.

        Parameters
        ----------
        value : *
            Any value to append.
        """
        import apysc as ap
        from apysc._type import value_util
        value_str: str = value_util.get_value_str_for_expression(
            value=value)
        expression: str = (
            f'{self.variable_name}.push({value_str});'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def extend(self, other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Concatenate argument array to this one. This interface
        positions the argument array's values after this array
        values. This method is similar to the concat method.
        Still, there is a difference in whether updating the same
        variable (extend) or returned as a different variable (concat).

        Parameters
        ----------
        other_arr : Array or list or tuple
            Other array-like values to concatenate.

        References
        ----------
        - Array class extend and concat interfaces document
            - https://simon-ritchie.github.io/apysc/array_extend_and_concat.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.extend([4, 5, 6])
        >>> arr
        Array([1, 2, 3, 4, 5, 6])
        """
        self._validate_acceptable_value_type(value=other_arr)
        if isinstance(other_arr, Array):
            self._value.extend(other_arr.value)  # type: ignore
        else:
            self._value.extend(list(other_arr))  # type: ignore[arg-type]
        self._append_extend_expression(other_arr=other_arr)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_extend_expression(
            self, *, other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Append an `extend` method expression.

        Parameters
        ----------
        other_arr : Array or list or tuple
            The other array-like value to concatenate.
        """
        import apysc as ap
        from apysc._type import value_util
        value_str: str = value_util.get_value_str_for_expression(
            value=other_arr)
        expression: str = (
            f'{self.variable_name} = {self.variable_name}'
            f'.concat({value_str});')
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def concat(
            self, other_arr: Union[List[T], tuple, 'Array']) -> 'Array':
        """
        Concatenate argument array to this one. This interface
        positions the argument array's values after this array
        values. This method is similar to extend method,
        but there is a difference in whether updating
        the same variable (extend) or returned as a different
        variable (concat).

        Parameters
        ----------
        other_arr : Array or list or tuple
            Other array-like values to concatenate.

        Returns
        -------
        concatenated : Array
            Concatenated array value.

        References
        ----------
        - Array class extend and concat interfaces document
            - https://simon-ritchie.github.io/apysc/array_extend_and_concat.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr = arr.concat([4, 5, 6])
        >>> arr
        Array([1, 2, 3, 4, 5, 6])
        """
        concatenated: Array = self._copy()
        concatenated.extend(other_arr)
        self._append_concat_expression(
            concatenated=concatenated, other_arr=other_arr)
        return concatenated

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_concat_expression(
            self, *, concatenated: VariableNameInterface,
            other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Append the `concat` method expression.

        Parameters
        ----------
        concatenated : Array
            Concatenated array value.
        other_arr : Array or list or tuple
            The other array-like value to concatenate.
        """
        import apysc as ap
        from apysc._type import value_util
        value_str: str = value_util.get_value_str_for_expression(
            value=other_arr)
        expression: str = (
            f'var {concatenated.variable_name} = '
            f'{self.variable_name}.concat({value_str});'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def insert(
            self, index: Union[int, Int], value: T) -> None:
        """
        Insert value to this array at a specified index.
        This interface behaves the same `insert_at` method.

        Parameters
        ----------
        index : Int or int
            Index to append value.
        value : *
            Any value to append.

        References
        ----------
        - Array class insert and insert_at interfaces document
            - https://simon-ritchie.github.io/apysc/array_insert_and_insert_at.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3])
        >>> arr.insert(index=1, value=2)
        >>> arr
        Array([1, 2, 3])
        """
        import apysc as ap
        from apysc._validation import number_validation
        number_validation.validate_integer(integer=index)
        if isinstance(index, ap.Int):
            index_: int = int(index.value)
        else:
            index_ = int(index)
        value_: Any
        if isinstance(value, ap.Int):
            value_ = int(value.value)
        else:
            value_ = value
        self._value.insert(index_, value_)
        self._append_insert_expression(index=index, value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def insert_at(self, index: Union[int, Int], value: T) -> None:
        """
        Insert value to this array at a specified index.
        This interface behaves the same `insert` method.

        Parameters
        ----------
        index : Int or int
            Index to append value.
        value : *
            Any value to append.

        References
        ----------
        - Array class insert and insert_at interfaces document
            - https://simon-ritchie.github.io/apysc/array_insert_and_insert_at.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3])
        >>> arr.insert_at(index=1, value=2)
        >>> arr
        Array([1, 2, 3])
        """
        self.insert(index=index, value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_insert_expression(
            self, *, index: Union[int, Int], value: T) -> None:
        """
        Append insert method expression.

        Parameters
        ----------
        index : Int or int
            Index to append value.
        value : *
            Any value to append.
        """
        import apysc as ap
        from apysc._type import value_util
        value_str: str = value_util.get_value_str_for_expression(
            value=value)
        index_str: str = value_util.get_value_str_for_expression(
            value=index)
        expression: str = (
            f'{self.variable_name}.splice({index_str}, 0, {value_str});'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def pop(self) -> T:
        """
        Remove this array's last value and return it.

        Returns
        -------
        value : *
            Removed value.

        References
        ----------
        - Array class pop interface document
            - https://simon-ritchie.github.io/apysc/array_pop.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> popped_val: int = arr.pop()
        >>> popped_val
        3

        >>> arr
        Array([1, 2])
        """
        value: T = self._value.pop()
        self._append_pop_expression(value=value)
        return value

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_pop_expression(self, *, value: T) -> None:
        """
        Append pop method expression.

        Parameters
        ----------
        value : *
            Removed value.
        """
        import apysc as ap
        expression: str = f'{self.variable_name}.pop();'
        if isinstance(value, VariableNameInterface):
            expression = f'{value.variable_name} = {expression}'
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def remove(self, value: T) -> None:
        """
        Remove a specified value from this array.

        Parameters
        ----------
        value : Any
            Value to remove.

        References
        ----------
        - Array class remove and remove_at interfaces document
            - https://simon-ritchie.github.io/apysc/array_remove_and_remove_at.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3, 5])
        >>> arr.remove(3)
        >>> arr
        Array([1, 5])
        """
        self._value.remove(value)
        self._append_remove_expression(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_remove_expression(self, *, value: T) -> None:
        """
        Append a `remove` method expression.

        Parameters
        ----------
        value : Any
            Value to remove.
        """
        import apysc as ap
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._type import value_util
        index_var_name: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.INDEX)
        value_str: str = value_util.get_value_str_for_expression(
            value=value)
        expression: str = (
            f'var {index_var_name} = _.indexOf'
            f'({self.variable_name}, {value_str});'
            f'\n{self.variable_name}.splice({index_var_name}, 1);')
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def remove_at(self, index: Union[int, Int]) -> None:
        """
        Remove a specified index value from this array.

        Parameters
        ----------
        index : Int or int
            Index to remove value.

        References
        ----------
        - Array class remove and remove_at interfaces document
            - https://simon-ritchie.github.io/apysc/array_remove_and_remove_at.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.remove_at(1)
        >>> arr
        Array([1, 3])
        """
        import apysc as ap
        self._validate_index_type_is_int(index=index)
        if isinstance(index, ap.Int):
            index_: int = int(index.value)
        else:
            index_ = int(index)
        if index_ in self._value:
            del self._value[index_]
        self._append_remove_at_expression(index=index)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_remove_at_expression(
            self, *, index: Union[int, Int]) -> None:
        """
        Append remove_at method expression.

        Parameters
        ----------
        index : Int or int
            Index to remove value.
        """
        import apysc as ap
        from apysc._type import value_util
        index_str: str = value_util.get_value_str_for_expression(
            value=index)
        expression: str = (
            f'{self.variable_name}.splice({index_str}, 1);'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def reverse(self) -> None:
        """
        Reverse this array in place.

        References
        ----------
        - Array class reverse interface document
            - https://simon-ritchie.github.io/apysc/array_reverse.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.reverse()
        >>> arr
        Array([3, 2, 1])
        """
        self._value.reverse()
        self._append_reverse_expression()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_reverse_expression(self) -> None:
        """
        Append reverse method expression.
        """
        import apysc as ap
        expression: str = (
            f'{self.variable_name}.reverse();'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def sort(self, *, ascending: bool = True) -> None:
        """
        Sort this array in place.

        Parameters
        ----------
        ascending : bool, default True
            Sort by ascending or not. If False is specified,
            this interface sorts values descending.

        References
        ----------
        - Array class sort interface document
            - https://simon-ritchie.github.io/apysc/array_sort.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([3, 5, 1, 4, 2])
        >>> arr.sort()
        >>> arr
        Array([1, 2, 3, 4, 5])

        >>> arr.sort(ascending=False)
        >>> arr
        Array([5, 4, 3, 2, 1])
        """
        self._value.sort()
        self._append_sort_expression()
        if not ascending:
            self.reverse()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_sort_expression(self) -> None:
        """
        Append sort method expression.
        """
        import apysc as ap
        expression: str = (
            f'{self.variable_name}.sort();'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def slice(
            self,
            *,
            start: Optional[Union[int, Int]] = None,
            end: Optional[Union[int, Int]] = None) -> 'Array':
        """
        Slice this array by specified start and end indexes.

        Parameters
        ----------
        start : Int or int or None, default None
            Slicing start index.
        end : Int or int or None, default None
            Slicing end index (a result array does not contain
            this index).

        Returns
        -------
        sliced_arr : Array
            Sliced array.

        References
        ----------
        - Array class slice interface document
            - https://simon-ritchie.github.io/apysc/array_slice.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3, 4])
        >>> arr.slice(start=1, end=3)
        Array([2, 3])

        >>> arr.slice(start=1)
        Array([2, 3, 4])

        >>> arr.slice(end=2)
        Array([1, 2])
        """
        import apysc as ap
        if isinstance(start, ap.Int):
            start_: Optional[int] = int(start.value)
        else:
            start_ = start  # type: ignore[assignment]
        if isinstance(end, ap.Int):
            end_: Optional[int] = int(end.value)
        else:
            end_ = end  # type: ignore[assignment]
        sliced_arr: Array = self._copy()
        sliced_arr._value = self._value[slice(start_, end_)]
        self._append_slice_expression(
            sliced_arr=sliced_arr, start=start, end=end)
        return sliced_arr

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_slice_expression(
            self,
            *,
            sliced_arr: VariableNameInterface,
            start: Optional[Union[int, Int]],
            end: Optional[Union[int, Int]]) -> None:
        """
        Append slice method expression.

        Parameters
        ----------
        sliced_arr : Array
            Sliced array value.
        start : Int or int or None
            Slicing start index.
        end : Int or int or None
            Slicing end index.
        """
        import apysc as ap
        if start is None:
            start = 0
        expression: str = (
            f'var {sliced_arr.variable_name} = '
            f'{self.variable_name}.slice('
            f'{start}'
        )
        if end is not None:
            expression += f', {end}'
        expression += ');'
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __getitem__(self, index: Union[int, Int]) -> T:
        """
        Get a specified index single value.

        Parameters
        ----------
        index : Int or int
            Array's index to get value. Currently not supported tuple
            value (e.g., slicing).

        Returns
        -------
        value : *
            Specified index's value.

        Raises
        ------
        ValueError
            If specified index type is not the `Int` or `int` type.
        """
        import apysc as ap
        self._validate_index_type_is_int(index=index)
        index_: int = self._get_builtin_int_from_index(index=index)
        value: Any
        if len(self._value) <= index:
            value = ap.AnyValue(None)
        else:
            value = self._value[index_]
        self._append_getitem_expression(index=index, value=value)
        return value

    def _get_builtin_int_from_index(
            self, *, index: Union[int, Int]) -> int:
        """
        Get Python built-in integer from index value.

        Parameters
        ----------
        index : Int or int
            Specified an array's index.

        Returns
        -------
        builtin_int_index : int
            Python builtin integer index value.
        """
        import apysc as ap
        if isinstance(index, ap.Int):
            return int(index.value)
        return int(index)

    def _validate_index_type_is_int(
            self, *, index: Union[int, Int]) -> None:
        """
        Validate whether an index value type is an int (or Int)
        type or not.

        Parameters
        ----------
        index : Int or int
            Index value to check.

        Raises
        ------
        ValueError
            If index type is not int or Int type.
        """
        if isinstance(index, (int, Int)):
            return
        raise ValueError(
            'Currently indexing is only supported int or Int types.'
            ' If you need to slice array please use slice method.')

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_getitem_expression(
            self, *, index: Union[int, Int],
            value: T) -> None:
        """
        Append __getitem__ expression.

        Parameters
        ----------
        index : Int or int
            Array's index to get value.
        value : *
            Specified index's value.
        """
        import apysc as ap
        from apysc._type import value_util
        value_: VariableNameInterface
        if not isinstance(value, VariableNameInterface):
            value_ = ap.AnyValue(None)
        else:
            value_ = value
        index_str: str = value_util.get_value_str_for_expression(
            value=index)
        expression: str = (
            f'var {value_.variable_name} = '
            f'{self.variable_name}[{index_str}];'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __setitem__(self, index: Union[int, Int], value: T) -> None:
        """
        Set value to a specified index.

        Parameters
        ----------
        index : Int or int
            Array's index to set value. Currently not supported tuple
            value (e.g., slicing).
        value : *
            Any value to set.

        Raises
        ------
        ValueError
            If a specified index type is not an `Int` or an
            `int` type.
        """
        self._validate_index_type_is_int(index=index)
        index_: int = self._get_builtin_int_from_index(index=index)
        self._value[index_] = value
        self._append_setitem_expression(index=index, value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_setitem_expression(
            self, *, index: Union[int, Int], value: T) -> None:
        """
        Append __setitem__ method expression.

        Parameters
        ----------
        index : Int or int
            Array's index to set value. Currently not supported tuple
            value (e.g., slicing).
        value : *
            Any value to set.
        """
        import apysc as ap
        from apysc._type import value_util
        index_str: str = value_util.get_value_str_for_expression(
            value=index)
        value_str: str = value_util.get_value_str_for_expression(
            value=value)
        expression: str = (
            f'{self.variable_name}[{index_str}] = '
            f'{value_str};'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __delitem__(self, index: Union[int, Int]) -> None:
        """
        Delete a specified index value from this array value.

        Parameters
        ----------
        index : Int or int
            Array's index to delete. Currently not supported tuple
            value (e.g., slicing).

        Raises
        ------
        ValueError
            - ValueError: If specified index type is not the
            `Int` or `int`.
        """
        self.remove_at(index=index)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def length(self) -> Int:
        """
        Get length of this array.

        Returns
        -------
        length : Int
            This array's length.

        References
        ----------
        - Array class length interface document
            - https://simon-ritchie.github.io/apysc/array_length.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.length
        Int(3)
        """
        import apysc as ap
        length: ap.Int = ap.Int(len(self._value))
        self._append_length_expression(length=length)
        return length

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_length_expression(self, *, length: Int) -> None:
        """
        Append length method expression.

        Parameters
        ----------
        length : Int
            Created length Int variable.
        """
        import apysc as ap
        expression: str = (
            f'{length.variable_name} = {self.variable_name}.length;'
        )
        ap.append_js_expression(expression=expression)

    def __len__(self) -> None:
        """
        This method is disabled and can't use from an
        Array instance.
        """
        raise Exception(
            'Array instance can\'t apply len function.'
            ' Please use length property instead.')

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def join(self, sep: Union[str, String]) -> String:
        """
        Join this array value with a specified separator string.

        Parameters
        ----------
        sep : String or str
            Separator string.

        Returns
        -------
        joined : String
            Joined string.

        References
        ----------
        - Array class join interface document
            - https://simon-ritchie.github.io/apysc/array_join.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.join(sep=', ')
        String('1, 2, 3')
        """
        import apysc as ap
        if isinstance(sep, ap.String):
            sep_: str = sep._value
        else:
            sep_ = str(sep)
        values_: List[Any] = [str(value) for value in self._value]
        joined: ap.String = ap.String(sep_.join(values_))
        self._append_join_expression(joined=joined, sep=sep)
        return joined

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_join_expression(
            self, *, joined: String, sep: Union[str, String]) -> None:
        """
        Append a `join` method expression.

        Parameters
        ----------
        joined : String
            Joined string.
        sep : String or str
            Separator string.
        """
        import apysc as ap
        from apysc._type import value_util
        sep_str: str = value_util.get_value_str_for_expression(value=sep)
        expression: str = (
            f'{joined.variable_name} = {self.variable_name}'
            f'.join({sep_str});'
        )
        ap.append_js_expression(expression=expression)

    def __str__(self) -> str:
        """
        String conversion method.

        Returns
        -------
        string : str
            Converted value string.
        """
        if not hasattr(self, '_value'):
            return '[]'
        return str(self._value)

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Array([])'
        else:
            repr_str = f'Array({repr(self._value)})'
        return repr_str

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def index_of(self, value: T) -> Int:
        """
        Search specified value's index and return it.

        Parameters
        ----------
        value : *
            Any value to search.

        Returns
        -------
        index : Int
            Found position of index. If this array does not
            contain a value, this interface returns -1.

        References
        ----------
        - Array class index_of interface document
            - https://simon-ritchie.github.io/apysc/array_index_of.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3, 5])
        >>> arr.index_of(3)
        Int(1)
        """
        import apysc as ap
        index: ap.Int = ap.Int(-1)
        try:
            index_: int = self._value.index(value)
        except Exception:
            index_ = -1
        index._value = index_
        self._append_index_of_expression(index=index, value=value)
        return index

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_index_of_expression(
            self, *, index: Int, value: T) -> None:
        """
        Append index_of method expression.

        Parameters
        ----------
        index : Int
            Found position of index. If an array does not
            contain a value, this interface returns -1.
        value : *
            Any value to find.
        """
        import apysc as ap
        from apysc._type import value_util
        value_str: str = value_util.get_value_str_for_expression(
            value=value)
        expression: str = (
            f'{index.variable_name} = {self.variable_name}'
            f'.indexOf({value_str});'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : *
            The other value to compare. A list or an Array
            type is acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        result: ap.Boolean
        if isinstance(other, Array):
            result = ap.Boolean(self._value == other._value)
        else:
            result = ap.Boolean(self._value == other)
            other = self._convert_other_val_to_array(other=other)
        if isinstance(other, VariableNameInterface):
            self._append_eq_expression(result=result, other=other)
        return result

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _convert_other_val_to_array(self, *, other: Any) -> Any:
        """
        If a comparison's other value is list value, then convert
        it to Array instance.

        Parameters
        ----------
        other : *
            The other value to compare.

        Returns
        -------
        converted_val : *
            Converted value. If the other value is a list,
            this value becomes Array type. Otherwise, this
            interface returns its value directly (skips conversion).
        """
        if isinstance(other, list):
            return Array(other)
        return other

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_eq_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append an __eq__ expression.

        Parameters
        ----------
        result : Boolean
            A result boolean value.
        other : Array
            The other value to compare.
        """
        import apysc as ap
        expression: str = (
            f'{result.variable_name} = '
            f'_.isEqual({self.variable_name}, {other.variable_name});'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def __ne__(self, other: Any) -> Any:
        """
        Not equal comparison method.

        Parameters
        ----------
        other : *
            The other value to compare. A list or an Array
            type is acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        other = self._convert_other_val_to_array(other=other)
        result: ap.Boolean = self == other
        result = result.not_
        if isinstance(other, VariableNameInterface):
            self._append_ne_expression(result=result, other=other)
        return result

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Array')
    def _append_ne_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append a __ne__ expression.

        Parameters
        ----------
        result : Boolean
            A result boolean value.
        other : Array
            The other value to compare.
        """
        import apysc as ap
        expression: str = (
            f'{result.variable_name} = '
            f'!_.isEqual({self.variable_name}, {other.variable_name});'
        )
        ap.append_js_expression(expression=expression)

    def __bool__(self) -> bool:
        """
        Get a boolean value whether this array value is empty
        or not.

        Returns
        -------
        result : bool
            If this array value is empty, this interface returns True.
        """
        return bool(self._value)

    _value_snapshots: Dict[str, List[T]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=[*self._value], snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: int_and_number.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Int and Number':
    '',

    'This page explains the `Int` and `Number` classes.\n\nBefore reading on, maybe it is helpful to read the following page:\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## Int class':
    '',

    'The `Int` class is the apysc integer type. It can accept numeric values at the constructor, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nassert int_1 == 10\n```':  # noqa
    '',

    '':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nint_2: ap.Int = ap.Int(int_1)\n```':  # noqa
    '',

    '':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\nint_2: ap.Int = ap.Int(int_1)\nint_2 += 15\nassert int_2 == 25\n```':  # noqa
    '',

    'If you specify a float value to the constructor argument, then the `Int` class floor a value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10.5)\nassert int_1 == 10\n```':  # noqa
    '',

    '':
    '',

    '## Number class':
    '',

    'The ``Number`` class is the apysc float type. It can accept numeric values at the constructor, same as `Int`:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nnumber_1: ap.Number = ap.Number(10.5)\nassert number_1 == 10.5\n\nnumber_2: ap.Number = ap.Number(number_1)\nnumber_2 += 10.5\nassert number_2 == 21\n```':  # noqa
    '',

    '':
    '',

    '## Note for the Float class alias':
    '',

    'The `Float` class is the alias of the `Number` class. It behaves the same as the `Number` class. Maybe a Python developer is familiar with its name rather than the `Number`\\. On the other hand, the `Number` is more common in JavaScript than the `Number`\\.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nassert ap.Number == ap.Float\nassert ap.Number(10.5) == ap.Float(10.5)\n```':  # noqa
    '',

    '':
    '',

    '## Int and Number classes basic interfaces':
    '',

    'The `Int` and `Number` classes have the same interfaces. For more details, please see:\n\n- [Int and Number classes basic arithmetic operations](int_and_number_arithmetic_operations.md)\n- [Int and Number classes basic comparison operations](int_and_number_comparison_operations.md)\n- [Funcdamental data classes common value interface](fundamental_data_classes_value_interface.md)':  # noqa
    '',

    '## Int class constructor API':
    '',

    '<!-- Docstring: apysc._type.int.Int.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, value:Union[int, float, apysc._type.number_value_interface.NumberValueInterface]) -> None`<hr>\n\n**[Interface summary]** Integer class for apysc library.<hr>\n\n**[Parameters]**\n\n- `value`: int or float or Int or Number\n  - Initial integer value. If the `float` or `Number` value is specified, this class casts it to an integer.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> int_val: ap.Int = ap.Int(10)\n>>> int_val\nInt(10)\n\n>>> int_val == 10\nBoolean(True)\n\n>>> int_val == ap.Int(10)\nBoolean(True)\n\n>>> int_val >= 10\nBoolean(True)\n\n>>> int_val += 10\n>>> int_val\nInt(20)\n\n>>> int_val = ap.Int(10.5)\n>>> int_val\nInt(10)\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Int and Number common arithmetic operations document](https://simon-ritchie.github.io/apysc/int_and_number_arithmetic_operations.html)\n- [Int and Number common comparison operations document](https://simon-ritchie.github.io/apysc/int_and_number_comparison_operations.html)':  # noqa
    '',

    '## Number class constructor API':
    '',

    '<!-- Docstring: apysc._type.number.Number.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, value:Union[int, float, apysc._type.number_value_interface.NumberValueInterface]) -> None`<hr>\n\n**[Interface summary]** Floating point number class for apysc library.<hr>\n\n**[Parameters]**\n\n- `value`: int or float or Int or Number\n  - Initial floating point number value. This class casts it to float if you specify int or Int value.\n\n<hr>\n\n**[Notes]**\n\nThe `Float` class is the alias of the Number, and it behaves the same as the Number class.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> number: ap.Number = ap.Number(10.5)\n>>> number\nNumber(10.5)\n\n>>> number == 10.5\nBoolean(True)\n\n>>> number == ap.Number(10.5)\nBoolean(True)\n\n>>> number >= 10.5\nBoolean(True)\n\n>>> number += 10.3\n>>> number\nNumber(20.8)\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Int and Number common arithmetic operations document](https://simon-ritchie.github.io/apysc/int_and_number_arithmetic_operations.html)\n- [Int and Number common comparison operations document](https://simon-ritchie.github.io/apysc/int_and_number_comparison_operations.html)':  # noqa
    '',

    '## value property API':
    '',

    '<!-- Docstring: apysc._type.number_value_interface.NumberValueInterface.value -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a current number value.<hr>\n\n**[Returns]**\n\n- `value`: int or float\n  - Current number value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> int_val: ap.Int = ap.Int(10)\n>>> int_val.value\n10\n\n>>> int_val.value = 20\n>>> int_val.value\n20\n\n>>> int_val.value = ap.Int(30)\n>>> int_val.value\n30\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [apysc fundamental data classes value interface](https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html)':  # noqa
    '',

}



"""Interface class implementation for the relative value.
"""

from typing import Dict

from apysc._type.boolean import Boolean
from apysc._type.revert_interface import RevertInterface


class RelativeInterface(RevertInterface):

    _relative: Boolean

    def _initialize_relative_if_not_initialized(self) -> None:
        """
        Initialize the _relative attribute if this instance
        does not initialize it yet.
        """
        if hasattr(self, '_relative'):
            return
        self._relative = Boolean(False)

    @property
    def relative(self) -> Boolean:
        """
        Get a boolean value indicating whether a path data
        is relative or not.

        Returns
        -------
        relative : Boolean
            A boolean value indicating whether path data
            is relative or not.

        Examples
        --------
        >>> import apysc as ap
        >>> line_to: ap.PathLineTo = ap.PathLineTo(
        ...     x=50, y=50, relative=False)
        >>> line_to.relative = ap.Boolean(True)
        >>> line_to.relative
        Boolean(True)
        """
        self._initialize_relative_if_not_initialized()
        return self._relative._copy()

    @relative.setter
    def relative(self, value: Boolean) -> None:
        """
        Set a boolean value indicating whether path a path data
        is relative or not.

        Parameters
        ----------
        value : Boolean
            A boolean value indicating whether path data
            is relative or not.
        """
        self._initialize_relative_if_not_initialized()
        self._relative.value = value

    _relative_snapshots: Dict[str, bool]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_relative_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_relative_snapshots',
            value=self._relative._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_relative_if_not_initialized()
        self._relative._value = self._relative_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: array_length.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class length interface':
    '',

    'This page explains the `Array` class `length` property interface.':
    '',

    '## What interface is this?':
    '',

    'The `length` attribute interface returns a current array\'s values length.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `length` property has the only getter interface. The return value type is the `Int` type.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3, 4])\nlength: ap.Int = arr.length\nassert length == 4\n```':  # noqa
    '',

    '':
    '',

    '## Notes of the len() function':
    '',

    'The `Array` class is not supported the Python built-in `len()` function, and its function raises an exception. Please use the `length` property instead.':  # noqa
    '',

    '```py\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3, 4])\nlen(arr)\n```':  # noqa
    '',

    '':
    '',

    '```\nException: Array instance can\'t apply len function. Please use length property instead.\n```':  # noqa
    '',

    '':
    '',

    '## length property API':
    '',

    '<!-- Docstring: apysc._type.array.Array.length -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get length of this array.<hr>\n\n**[Returns]**\n\n- `length`: Int\n  - This array\'s length.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 2, 3])\n>>> arr.length\nInt(3)\n```':  # noqa
    '',

}



"""Class implementation for the timer.
"""

from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._event import timer_event
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._event.handler import HandlerData
from apysc._html.debug_mode import add_debug_info_setting
from apysc._time.fps import FPS
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.number_value_interface import NumberValueInterface
from apysc._type.variable_name_interface import VariableNameInterface

_O1 = TypeVar('_O1')
_O2 = TypeVar('_O2')
_Handler = Callable[['timer_event.TimerEvent', _O1], None]


class Timer(VariableNameInterface, CustomEventInterface):
    """
    Timer class to handle function calling at regular intervals.

    References
    ----------
    - Timer document
        - https://simon-ritchie.github.io/apysc/timer.html
    - TimerEvent class document
        - https://simon-ritchie.github.io/apysc/timer_event.html
    - Timer class delay setting document
        - https://simon-ritchie.github.io/apysc/timer_delay.html
    - FPS enum document
        - https://simon-ritchie.github.io/apysc/fps.html
    - Timer class repeat_count setting
        - https://simon-ritchie.github.io/apysc/timer_repeat_count.html

    Examples
    --------
    >>> from typing_extensions import TypedDict
    >>> import apysc as ap
    >>> class RectOptions(TypedDict):
    ...     rectangle: ap.Rectangle
    >>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:
    ...     rectangle: ap.Rectangle = options['rectangle']
    ...     rectangle.x += 1
    >>> def on_timer_complete(e: ap.TimerEvent, options: dict) -> None:
    ...     ap.trace('Timer completed!')
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50)
    >>> options: RectOptions = {'rectangle': rectangle}
    >>> timer: ap.Timer = ap.Timer(
    ...     on_timer, delay=ap.FPS.FPS_60, repeat_count=50,
    ...     options=options)
    >>> _ = timer.timer_complete(on_timer_complete)
    >>> timer.start()
    """

    _delay: Number
    _repeat_count: Int
    _current_count: Int
    _handler_data: HandlerData
    _handler_name: str
    _running: Boolean

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def __init__(
            self,
            handler: _Handler[_O1],
            delay: Union[int, float, NumberValueInterface, FPS],
            *,
            repeat_count: Union[int, Int] = 0,
            options: Optional[_O1] = None) -> None:
        """
        Timer class to handle function calling at regular intervals.

        Parameters
        ----------
        handler : _Handler
            A handler would be called at regular intervals.
        delay : Int or int or Number or float or FPS
            A delay between each `Handler` calling in a millisecond
            or FPS value. If an `FPS` value is specified, this
            value becomes a millisecond calculated with that
            FPS value (e.g., if the `FPS_60` value is specified,
            then `delay` becomes 16.6666667).
        repeat_count : Int or int
            Max count of a `Handler`'s calling. A timer stops
            if the `Handler`'s calling count has reached this value.
            If 0 is specified, then a timer loops forever.
        options : dict or None, default None
            Optional arguments dictionary to pass a `Handler` callable.

        References
        ----------
        - Timer document
            - https://simon-ritchie.github.io/apysc/timer.html
        - TimerEvent class document
            - https://simon-ritchie.github.io/apysc/timer_event.html
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html
        - FPS enum document
            - https://simon-ritchie.github.io/apysc/fps.html
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html
        - About the handler options' type document
            - https://simon-ritchie.github.io/apysc/about_handler_options_type.html  # noqa

        Examples
        --------
        >>> from typing_extensions import TypedDict
        >>> import apysc as ap
        >>> class RectOptions(TypedDict):
        ...     rectangle: ap.Rectangle
        >>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:
        ...     rectangle: ap.Rectangle = options['rectangle']
        ...     rectangle.x += 1
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> options: RectOptions = {'rectangle': rectangle}
        >>> _ = ap.Timer(
        ...     on_timer, delay=ap.FPS.FPS_60, options=options).start()
        """
        import apysc as ap
        from apysc._event.handler import get_handler_name
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._expression.event_handler_scope import \
            TemporaryNotHandlerScope
        from apysc._validation import number_validation
        from apysc._validation.handler_options_validation import \
            validate_options_type
        with TemporaryNotHandlerScope():
            validate_options_type(options=options)
            self.variable_name = \
                expression_variables_util.get_next_variable_name(
                    type_name=var_names.TIMER)
            self._handler_name = get_handler_name(
                handler=handler, instance=self)
            handler = self._wrap_handler(handler=handler)
            self._handler: _Handler[_O1] = handler
            delay_: ap.Number = self._convert_delay_to_number(
                delay=delay)
            number_validation.validate_num_is_gte_zero(num=delay_)
            self._delay = delay_
            if isinstance(repeat_count, ap.Int):
                repeat_count_: ap.Int = repeat_count
            else:
                repeat_count_ = ap.Int(repeat_count)
            number_validation.validate_num_is_gte_zero(num=repeat_count)
            self._repeat_count = repeat_count_
            self._running = ap.Boolean(False)
            self._current_count = ap.Int(0)
            if options is None:
                options = {}  # type: ignore
            self._handler_data = {  # type: ignore
                'handler': self._handler,  # type: ignore
                'options': options,
            }
            ap.append_js_expression(
                expression=f'var {self.variable_name};')

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def _convert_delay_to_number(
            self, *,
            delay: Union[int, float, NumberValueInterface, FPS]) -> Number:
        """
        Convert each type of delay value to a `Number` value.

        Parameters
        ----------
        delay : int or float or Int or Number or FPS
            A delay between each handler's calling in milliseconds
            or FPS value.

        Returns
        -------
        delay_ : Number
            Converted delay value.
        """
        import apysc as ap
        from apysc._time.fps import FPSDefinition
        if isinstance(delay, FPS):
            fps_definition: FPSDefinition = delay.value
            delay = fps_definition.milisecond_intervals
        if not isinstance(delay, ap.Number):
            delay_: ap.Number = ap.Number(delay)
        else:
            delay_ = delay
        return delay_

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def delay(self) -> Number:
        """
        Get a delay value.

        Returns
        -------
        delay : Number
            A delay value of each `Handler` calling in milliseconds.

        References
        ----------
        - Timer class delay setting document
            - https://simon-ritchie.github.io/apysc/timer_delay.html

        Examples
        --------
        >>> import apysc as ap
        >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
        ...     pass
        >>> timer: ap.Timer = ap.Timer(
        ...     on_timer, delay=33.3, repeat_count=50)
        >>> timer.delay
        Number(33.3)
        """
        from apysc._type import value_util
        return value_util.get_copy(value=self._delay)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def repeat_count(self) -> Int:
        """
        Get a max count value of a handler's calling.

        Returns
        -------
        repeat_count : Int
            Max count of a handler's calling.
            If this value is 0, then a timer loop forever.

        References
        ----------
        - Timer class repeat_count setting
            - https://simon-ritchie.github.io/apysc/timer_repeat_count.html

        Examples
        --------
        >>> import apysc as ap
        >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
        ...     pass
        >>> timer: ap.Timer = ap.Timer(
        ...     on_timer, delay=33.3, repeat_count=50)
        >>> timer.repeat_count
        Int(50)
        """
        from apysc._type import value_util
        return value_util.get_copy(value=self._repeat_count)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def running(self) -> Boolean:
        """
        Get a boolean value whether this timer is running or not.

        Returns
        -------
        running : Boolean.
            A boolean value whether this timer is running or not.

        Examples
        --------
        >>> import apysc as ap
        >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
        ...     pass
        >>> timer: ap.Timer = ap.Timer(
        ...     on_timer, delay=33.3, repeat_count=50)
        >>> timer.running
        Boolean(False)

        >>> timer.start()
        >>> timer.running
        Boolean(True)
        """
        from apysc._type import value_util
        return value_util.get_copy(value=self._running)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def current_count(self) -> Int:
        """
        Get a current timer count.

        Returns
        -------
        current_count : Int
            A current timer count.

        Examples
        --------
        >>> import apysc as ap
        >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
        ...     timer: ap.Timer = e.this
        ...     ap.trace(timer.current_count)
        >>> _ = ap.Timer(
        ...     on_timer, delay=33.3, repeat_count=50).start()
        """
        from apysc._type import value_util
        return value_util.get_copy(value=self._current_count)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def start(self) -> None:
        """
        Start this timer.

        References
        ----------
        - Timer class start and stop interfaces
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html

        Examples
        --------
        >>> import apysc as ap
        >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
        ...     pass
        >>> _ = ap.Timer(
        ...     on_timer, delay=33.3, repeat_count=50).start()
        """
        import apysc as ap
        from apysc._event import handler_circular_calling_util
        from apysc._event.handler import append_handler_expression
        from apysc._expression import expression_data_util
        from apysc._type import value_util
        delay_val_str: str = value_util.get_value_str_for_expression(
            value=self._delay)
        is_handler_circular_calling: bool = handler_circular_calling_util.\
            is_handler_circular_calling(handler_name=self._handler_name)
        if is_handler_circular_calling:
            handler_name: str = handler_circular_calling_util.\
                get_prev_handler_name(handler_name=self._handler_name)
            variable_name: str = handler_circular_calling_util.\
                get_prev_variable_name(handler_name=self._handler_name)
        else:
            handler_name = self._handler_name
            variable_name = self.variable_name
        expression: str = (
            f'\nif (_.isUndefined({variable_name})) {{'
            f'\n  {variable_name} = setInterval('
            f'{handler_name}, {delay_val_str});'
            '\n}'
        )
        expression_data_util.append_js_expression(expression=expression)

        e: ap.TimerEvent = ap.TimerEvent(this=self)
        append_handler_expression(
            handler_data=self._handler_data,
            handler_name=handler_name,
            e=e)
        self._running.value = True

    def _wrap_handler(self, *, handler: _Handler[_O1]) -> _Handler[_O1]:
        """
        Wrap a handler to update a current count value when
        it is called.

        Parameters
        ----------
        handler : _Handler
            Target handler.

        Returns
        -------
        wrapped : _Handler
            A wrapped handler.
        """
        from apysc._event.timer_event import TimerEvent

        def wrapped(e: TimerEvent, options: Any) -> None:
            """
            Wrapped handler.

            Parameters
            ----------
            e : TimerEvent
                Event instance.
            options : dict
                Optional arguments dictionary.
            """
            e.this._current_count += 1
            e.this._current_count._value = 0
            handler(e, options)
            self._append_count_branch_expression()

        return wrapped

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def _append_count_branch_expression(self) -> None:
        """
        Append the timer stopping expression by the counting.
        """
        from apysc._event.custom_event_type import CustomEventType
        from apysc._expression import expression_data_util
        from apysc._expression.indent_num import Indent
        from apysc._type import value_util
        current_count_val_str: str = value_util.\
            get_value_str_for_expression(value=self._current_count)
        repeat_count_val_str: str = value_util.\
            get_value_str_for_expression(value=self._repeat_count)
        expression: str = (
            f'if ({repeat_count_val_str} !== 0 '
            f'&& {current_count_val_str} === {repeat_count_val_str}) {{'
            f'\n{self._get_stop_expression(indent_num=1)}'
        )
        expression_data_util.append_js_expression(expression=expression)
        with Indent():
            self._running.value = False
            self.trigger_custom_event(
                custom_event_type=CustomEventType.TIMER_COMPLETE)
            self.unbind_custom_event_all(
                custom_event_type=CustomEventType.TIMER_COMPLETE)
            self._current_count.value = 0
        expression = '\n}'
        expression_data_util.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def stop(self) -> None:
        """
        Stop this timer.

        References
        ----------
        - Timer class start and stop interfaces document
            - https://simon-ritchie.github.io/apysc/timer_start_and_stop.html

        Examples
        --------
        >>> from typing_extensions import TypedDict
        >>> import apysc as ap
        >>> class RectOptions(TypedDict):
        ...     rectangle: ap.Rectangle
        >>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:
        ...     rectangle: ap.Rectangle = options['rectangle']
        ...     rectangle.x += 1
        ...     with ap.If(rectangle.x > 100):
        ...         timer: ap.Timer = e.this
        ...         timer.stop()
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> options: RectOptions = {'rectangle': rectangle}
        >>> _ = ap.Timer(
        ...     on_timer, delay=33.3, options=options).start()
        """
        from apysc._expression import expression_data_util
        self._running.value = False
        expression: str = self._get_stop_expression(indent_num=0)
        expression_data_util.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def _get_stop_expression(self, *, indent_num: int) -> str:
        """
        Get a stop interface expression string.

        Parameters
        ----------
        indent_num : int
            Indentation number to append.

        Returns
        -------
        expression : str
            Stop interface expression string.
        """
        from apysc._string import indent_util
        expression: str = (
            f'if (!_.isUndefined({self.variable_name})) {{'
            f'\n  clearInterval({self.variable_name});'
            f'\n  {self.variable_name} = undefined;'
            '\n}'
        )
        expression = indent_util.append_spaces_to_expression(
            expression=expression, indent_num=indent_num)
        return expression

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def reset(self) -> None:
        """
        Reset the timer count and stop this timer.

        References
        ----------
        - Timer class reset interface document
            - https://simon-ritchie.github.io/apysc/timer_reset.html

        Examples
        --------
        >>> from typing_extensions import TypedDict
        >>> import apysc as ap
        >>> class RectOptions(TypedDict):
        ...     rectangle: ap.Rectangle
        >>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:
        ...     rectangle: ap.Rectangle = options['rectangle']
        ...     rectangle.x += 1
        ...     with ap.If(rectangle.x > 100):
        ...         timer: ap.Timer = e.this
        ...         timer.reset()
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> options: RectOptions = {'rectangle': rectangle}
        >>> _ = ap.Timer(
        ...     on_timer, delay=33.3, options=options).start()
        """
        self.stop()
        self._current_count.value = 0

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Timer')
    def timer_complete(
            self, handler: _Handler[_O2], *,
            options: Optional[_O2] = None) -> str:
        """
        Add a timer complete event listener setting.

        Parameters
        ----------
        handler : _Handler
            A callable that a timer calls when complete.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.

        Returns
        -------
        name : str
            Handler's name.

        References
        ----------
        - Timer class timer_complete interface document
            - https://simon-ritchie.github.io/apysc/timer_complete.html
        - About the handler options’ type document
            - https://simon-ritchie.github.io/apysc/about_handler_options_type.html  # noqa

        Examples
        --------
        >>> from typing_extensions import TypedDict
        >>> import apysc as ap
        >>> class RectOptions(TypedDict):
        ...     rectangle: ap.Rectangle
        >>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:
        ...     rectangle: ap.Rectangle = options['rectangle']
        ...     rectangle.x += 1
        >>> def on_timer_complete(
        ...         e: ap.TimerEvent, options: RectOptions) -> None:
        ...     ap.trace('Timer completed!')
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> options: RectOptions = {'rectangle': rectangle}
        >>> timer: ap.Timer = ap.Timer(
        ...     on_timer, delay=33.3, options=options)
        >>> _ = timer.timer_complete(on_timer_complete)
        >>> timer.start()
        """
        import apysc as ap
        from apysc._event.custom_event_type import CustomEventType
        from apysc._validation.handler_options_validation import \
            validate_options_type
        validate_options_type(options=options)
        e: ap.TimerEvent = ap.TimerEvent(this=self)
        name: str = self.bind_custom_event(
            custom_event_type=CustomEventType.TIMER_COMPLETE,
            handler=handler,
            e=e,
            options=options)
        return name



"""Class implementation for the flip_y interface.
"""

from typing import Dict

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.boolean import Boolean
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class FlipYInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _flip_y: Boolean

    def _initialize_flip_y_if_not_initialized(self) -> None:
        """
        Initialize the _flip_y attribute if this interface
        does not initialize it yet.
        """
        if hasattr(self, '_flip_y'):
            return
        self._flip_y = Boolean(False)

        self._append_flip_y_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='FlipYInterface')
    def _append_flip_y_attr_linking_setting(self) -> None:
        """
        Append a flip-y attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._flip_y, attr_name='flip_y')
        self._append_attr_to_linking_stack(
            attr=self._flip_y, attr_name='flip_y')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='FlipYInterface')
    def flip_y(self) -> Boolean:
        """
        Get a boolean value whether the y-axis is flipping or not.

        Returns
        -------
        flip_y : Boolean
            A boolean value whether the y-axis is flipping or not.

        References
        ----------
        - GraphicsBase flip_x and flip_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_flip_interfaces.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> polygon: ap.Polygon = sprite.graphics.draw_polygon(
        ...     points=[
        ...         ap.Point2D(x=0, y=0),
        ...         ap.Point2D(x=50, y=0),
        ...         ap.Point2D(x=25, y=50),
        ...     ])
        >>> polygon.flip_y = ap.Boolean(True)
        >>> polygon.flip_y
        Boolean(True)
        """
        from apysc._type import value_util
        self._initialize_flip_y_if_not_initialized()
        return value_util.get_copy(value=self._flip_y)

    @flip_y.setter
    def flip_y(self, value: Boolean) -> None:
        """
        Update a y-axis flipping value.

        Parameters
        ----------
        value : Boolean
            Flipping value. If true, a y-axis will be flipped,
            otherwise it will be reset.

        References
        ----------
        - GraphicsBase flip_x and flip_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_flip_interfaces.html  # noqa
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='flip_y', args=[], kwargs={},
                module_name=__name__,
                class_name=FlipYInterface.__name__):
            import apysc as ap
            self._initialize_flip_y_if_not_initialized()
            before_value: ap.Boolean = self._flip_y
            self._flip_y = value
            self._append_flip_y_update_expression(before_value=before_value)

            self._append_flip_y_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='FlipYInterface')
    def _append_flip_y_update_expression(
            self, *, before_value: Boolean) -> None:
        """
        Append a y-axis flipping value updating expression.

        Parameters
        ----------
        before_value : Boolean
            Before updating flipping value.
        """
        import apysc as ap
        from apysc._display import flip_interface_helper
        self._initialize_flip_y_if_not_initialized()
        expression: str = flip_interface_helper.\
            make_flip_update_expression(
                before_value=before_value, after_value=self._flip_y,
                axis=flip_interface_helper.Axis.Y,
                interface_variable_name=self.variable_name)
        ap.append_js_expression(expression=expression)

    _flip_y_snapshots: Dict[str, bool]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_flip_y_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_flip_y_snapshots',
            value=self._flip_y._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._flip_y._value = self._flip_y_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: array_join.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class join interface':
    '',

    'This page explains the `Array` class `join` method interface.':
    '',

    '## What interface is this?':
    '',

    'The `join` method returns a joined `String` with the specified separator string.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `join` method requires the `sep` argument as the separator, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 2, 3])\njoined: ap.String = arr.join(sep=\',\')\nassert joined == \'1,2,3\'\n```':  # noqa
    '',

    '':
    '',

    '## join API':
    '',

    '<!-- Docstring: apysc._type.array.Array.join -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `join(self, sep:Union[str, apysc._type.string.String]) -> apysc._type.string.String`<hr>\n\n**[Interface summary]** Join this array value with a specified separator string.<hr>\n\n**[Parameters]**\n\n- `sep`: String or str\n  - Separator string.\n\n<hr>\n\n**[Returns]**\n\n- `joined`: String\n  - Joined string.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array([1, 2, 3])\n>>> arr.join(sep=\', \')\nString(\'1, 2, 3\')\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_scale_x_and_y_from_center.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# animation_scale_x_from_center and animation_scale_y_from_center interfaces':  # noqa
    '',

    'This page explains the `animation_scale_x_from_center` and `animation_scale_y_from_center` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `animation_scale_x_from_center` method interface creates an `ap.AnimationScaleXFromCenter` instance. You can animate x-direction\'s scale with it.\n\nSimilarly, the `animation_scale_y_from_center` method interface creates an `ap.AnimationScaleYFromCenter` instance. You can animate y-direction\'s scale with it.\n\nThese interfaces exist on a `GraphicsBase` subclass (that has the `scale_x_from_center` and `scale_y_from_center` interfaces), such as the `Rectangle` or `Circle`.':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets the scale-x (left-side rectangle) and scale-y (right-side rectangle) animation (from 1.0 to 2.0) with the `animation_scale_x_from_center` and `animation_scale_y_from_center` methods:':  # noqa
    '',

    '```py\n# runnable\nfrom enum import Enum\n\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\nDURATION: int = 1000\nEASING: ap.Easing = ap.Easing.EASE_OUT_QUINT\n\n\nclass Direction(Enum):\n    X = 1\n    Y = 2\n\n\nclass Options(TypedDict):\n    direction: Direction\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle], options: Options) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    SCALE: float = 1.0\n    if options[\'direction\'] == Direction.X:\n        rectangle.animation_scale_x_from_center(\n            scale_x_from_center=SCALE,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_2,\n            options=options,\n        ).start()\n    elif options[\'direction\'] == Direction.Y:\n        rectangle.animation_scale_y_from_center(\n            scale_y_from_center=SCALE,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_2,\n            options=options,\n        ).start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle], options: Options) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    SCALE: float = 2.0\n    if options[\'direction\'] == Direction.X:\n        rectangle.animation_scale_x_from_center(\n            scale_x_from_center=SCALE,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_1,\n            options=options,\n        ).start()\n    elif options[\'direction\'] == Direction.Y:\n        rectangle.animation_scale_y_from_center(\n            scale_y_from_center=SCALE,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_1,\n            options=options,\n        ).start()\n\n\nap.Stage(\n    stage_width=250, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nleft_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nright_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=150, y=50, width=50, height=50)\n\noptions: Options = {\'direction\': Direction.X}\nleft_rectangle.animation_scale_x_from_center(\n    scale_x_from_center=2.0,\n    duration=DURATION,\n    easing=EASING,\n).animation_complete(\n    on_animation_complete_1,\n    options=options,\n).start()\n\noptions = {\'direction\': Direction.Y}\nright_rectangle.animation_scale_y_from_center(\n    scale_y_from_center=2.0,\n    duration=DURATION,\n    easing=EASING,\n).animation_complete(\n    on_animation_complete_1,\n    options=options,\n).start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_scale_x_and_y_from_center_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_scale_x_and_y_from_center_basic_usage/index.html" width="250" height="150"></iframe>':  # noqa
    '',

    '## animation_scale_x_from_center API':
    '',

    '<!-- Docstring: apysc._animation.animation_scale_x_from_center_interface.AnimationScaleXFromCenterInterface.animation_scale_x_from_center -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_scale_x_from_center(self, scale_x_from_center:Union[float, apysc._type.number.Number], *, duration:Union[int, apysc._type.int.Int]=3000, delay:Union[int, apysc._type.int.Int]=0, easing:apysc._animation.easing.Easing=<Easing.LINEAR: \'function(x) {return x;}\'>) -> apysc._animation.animation_scale_x_from_center.AnimationScaleXFromCenter`<hr>\n\n**[Interface summary]** Set the scale-x from the center point animation setting.<hr>\n\n**[Parameters]**\n\n- `scale_x_from_center`: Number or float\n  - The final scale-x of the animation.\n- `duration`: Int or int, default 3000\n  - Milliseconds before an animation ends.\n- `delay`: Int or int, default 0\n  - Milliseconds before an animation starts.\n- `easing`: Easing, default Easing.LINEAR\n  - Easing setting.\n\n<hr>\n\n**[Returns]**\n\n- `animation_scale_x_from_center`: AnimationScaleXFromCenter\n  - Created animation setting instance.\n\n<hr>\n\n**[Notes]**\n\nTo start this animation, you need to call the `start` method of the returned instance.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_scale_x_from_center(\n...     scale_x_from_center=0.5,\n...     duration=1500,\n...     easing=ap.Easing.EASE_OUT_QUINT,\n... ).start()\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Animation interfaces duration setting document](https://simon-ritchie.github.io/apysc/animation_duration.html)\n- [Animation interfaces delay setting document](https://simon-ritchie.github.io/apysc/animation_delay.html)\n- [Each animation interface return value document](https://simon-ritchie.github.io/apysc/animation_return_value.html)\n- [Sequential animation setting document](https://simon-ritchie.github.io/apysc/sequential_animation.html)\n- [animation_parallel interface document](https://simon-ritchie.github.io/apysc/animation_parallel.html)\n- [Easing enum document](https://simon-ritchie.github.io/apysc/easing_enum.html)':  # noqa
    '',

    '## animation_scale_y_from_center API':
    '',

    '<!-- Docstring: apysc._animation.animation_scale_y_from_center_interface.AnimationScaleYFromCenterInterface.animation_scale_y_from_center -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_scale_y_from_center(self, scale_y_from_center:Union[float, apysc._type.number.Number], *, duration:Union[int, apysc._type.int.Int]=3000, delay:Union[int, apysc._type.int.Int]=0, easing:apysc._animation.easing.Easing=<Easing.LINEAR: \'function(x) {return x;}\'>) -> apysc._animation.animation_scale_y_from_center.AnimationScaleYFromCenter`<hr>\n\n**[Interface summary]** Set the scale-y from the center point animation setting.<hr>\n\n**[Parameters]**\n\n- `scale_y_from_center`: Number or float\n  - The final scale-y of the animation.\n- `duration`: Int or int, default 3000\n  - Milliseconds before an animation ends.\n- `delay`: Int or int, default 0\n  - Milliseconds before an animation starts.\n- `easing`: Easing, default Easing.LINEAR\n  - Easing setting.\n\n<hr>\n\n**[Returns]**\n\n- `animation_scale_y_from_center`: AnimationScaleYFromCenter\n  - Created animation setting instance.\n\n<hr>\n\n**[Notes]**\n\nTo start this animation, you need to call the `start` method of the returned instance.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_scale_y_from_center(\n...     scale_y_from_center=0.5,\n...     duration=1500,\n...     easing=ap.Easing.EASE_OUT_QUINT,\n... ).start()\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Animation interfaces duration setting document](https://simon-ritchie.github.io/apysc/animation_duration.html)\n- [Animation interfaces delay setting document](https://simon-ritchie.github.io/apysc/animation_delay.html)\n- [Each animation interface return value document](https://simon-ritchie.github.io/apysc/animation_return_value.html)\n- [Sequential animation setting document](https://simon-ritchie.github.io/apysc/sequential_animation.html)\n- [animation_parallel interface document](https://simon-ritchie.github.io/apysc/animation_parallel.html)\n- [Easing enum document](https://simon-ritchie.github.io/apysc/easing_enum.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: assert_equal_and_not_equal.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# assert_equal and assert_not_equal interfaces':
    '',

    'This page explains the `assert_equal` and `assert_not_equal` function interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `assert_equal` function interface asserts that two JavaScript values are equal. The `assert_not_equal` function interface asserts that the two JavaScript values are not equal.':  # noqa
    '',

    '## See also':
    '',

    '- [JavaScript assertion interface basic behavior](assertion_basic_behavior.md)':  # noqa
    '',

    '## Basic usage':
    '',

    'The `assert_equal` and `assert_not_equal` interfaces requires the `left` and `right` arguments. The `msg` argument is optional.\n\nIf the `left` value and `right` values are not the same, the `assert_equal` assertion fails and display an error message on the browser console:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1, msg=\'Values are not equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_equal_basic_usage/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_equal]\nLeft-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed: Values are not equal!\n```':  # noqa
    '',

    '<iframe src="static/assert_equal_basic_usage/index.html" width="0" height="0"></iframe>\n\nThe `assert_not_equal` interface has the same arguments, and if the `left` value and `right` value are the same values, an assertion fails:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\n\nint_1: ap.Int = ap.Int(10)\nap.assert_not_equal(left=10, right=int_1, msg=\'Values are equal!\')\n\nap.save_overall_html(\n    dest_dir_path=\'assert_not_equal_basic_usage/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_not_equal]\nRight-side variable name: i_11\nLeft value: 10 right value: 10\n...\nAssertion failed: Values are equal!\n```':  # noqa
    '',

    '<iframe src="static/assert_not_equal_basic_usage/index.html" width="0" height="0"></iframe>':  # noqa
    '',

    '## assert_equal API':
    '',

    '<!-- Docstring: apysc._console.assertion.assert_equal -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `assert_equal(left:Any, right:Any, *, msg:str=\'\') -> None`<hr>\n\n**[Interface summary]** JavaScript assertion interface for equal condition.<hr>\n\n**[Parameters]**\n\n- `left`: *\n  - Left-side value to compare.\n- `right`: *\n  - Right-side value to compare.\n- `msg`: str, optional\n  - Message to display when assertion failed.\n\n<hr>\n\n**[Notes]**\n\n ・If specified values are types of Array (or list), then this interface calls the assert_arrays_equal function instead. <br> ・If specified values are types of Dictionary (or dict), then this interface calls the assert_dicts_equal instead.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> int_1: ap.Int = ap.Int(10)\n>>> int_2: ap.Int = ap.Int(10)\n>>> ap.assert_equal(int_1, int_2)\n```':  # noqa
    '',

    '':
    '',

    '## assert_not_equal API':
    '',

    '<!-- Docstring: apysc._console.assertion.assert_not_equal -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `assert_not_equal(left:Any, right:Any, *, msg:str=\'\') -> None`<hr>\n\n**[Interface summary]** JavaScript assertion interface for not equal condition.<hr>\n\n**[Parameters]**\n\n- `left`: *\n  - Left-side value to compare.\n- `right`: *\n  - Right-side value to compare.\n- `msg`: str, optional\n  - Message to display when assertion failed.\n\n<hr>\n\n**[Notes]**\n\n ・If specified values are types of Array (or list), then this interface calls the assert_arrays_not_equal function instead. <br> ・If specified values are types of Dictionary (or dict), this interface calls the assert_dicts_not_equal function instead.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> int_1: ap.Int = ap.Int(10)\n>>> int_2: ap.Int = ap.Int(11)\n>>> ap.assert_not_equal(int_1, int_2)\n```':  # noqa
    '',

}



"""Line caps enum's definitions.
"""

from enum import Enum


class LineCaps(Enum):
    """
    The class for the line caps style type definitions.

    Attributes
    ----------
    BUTT
        Default style type. No line cap will be set.
    ROUND
        This style type will change line cap to round.
    SQUARE
        This style type will change line cap to square.
        This is similar to BUTT type, but the length of the line
        will be increased by the stroke-width (thickness) value.
    """
    BUTT = 'butt'
    ROUND = 'round'
    SQUARE = 'square'



"""Dash dot (1-dot chain) setting for a line.
"""

from typing import Union

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int


class LineDashDotSetting(Dictionary[str, Int]):
    """
    Dash dot (1-dot chain) setting for a line.

    References
    ----------
    - Graphics line_style interface document
        - https://simon-ritchie.github.io/apysc/graphics_line_style.html  # noqa

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.line_style(color='#fff', thickness=10)
    >>> line: ap.Line = sprite.graphics.draw_line(
    ...     x_start=50, y_start=50, x_end=150, y_end=50)
    >>> line.line_dash_dot_setting = ap.LineDashDotSetting(
    ...     dot_size=2, dash_size=5, space_size=3)
    >>> line.line_dash_dot_setting.dot_size
    Int(2)

    >>> line.line_dash_dot_setting.dash_size
    Int(5)

    >>> line.line_dash_dot_setting.space_size
    Int(3)
    """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineDashDotSetting')
    def __init__(
            self, dot_size: Union[int, Int],
            dash_size: Union[int, Int],
            space_size: Union[int, Int]) -> None:
        """
        Dash dot (1-dot chain) setting for a line.

        Parameters
        ----------
        dot_size : int or Int
            Dot size.
        dash_size : int or Int
            Dash size.
        space_size : int or Int
            Blank space size between dots and dashes.

        References
        ----------
        - Graphics line_style interface document
            - https://simon-ritchie.github.io/apysc/graphics_line_style.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_dash_dot_setting = ap.LineDashDotSetting(
        ...     dot_size=2, dash_size=5, space_size=3)
        >>> line.line_dash_dot_setting.dot_size
        Int(2)

        >>> line.line_dash_dot_setting.dash_size
        Int(5)

        >>> line.line_dash_dot_setting.space_size
        Int(3)
        """
        from apysc._converter.to_apysc_val_from_builtin import \
            get_copied_int_from_builtin_val
        from apysc._validation import number_validation
        number_validation.validate_nums_are_int_and_gt_zero(
            nums=[dot_size, dash_size, space_size])
        dot_size_: Int = get_copied_int_from_builtin_val(
            integer=dot_size)
        dash_size_: Int = get_copied_int_from_builtin_val(
            integer=dash_size)
        space_size_: Int = get_copied_int_from_builtin_val(
            integer=space_size)
        super(LineDashDotSetting, self).__init__({
            'dot_size': dot_size_,
            'dash_size': dash_size_,
            'space_size': space_size_,
        })

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineDashDotSetting')
    def dot_size(self) -> Int:
        """
        Get a dot size setting.

        Returns
        -------
        dot_size : Int
            Dot size setting.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_dash_dot_setting = ap.LineDashDotSetting(
        ...     dot_size=2, dash_size=5, space_size=3)
        >>> line.line_dash_dot_setting.dot_size
        Int(2)
        """
        return self['dot_size']

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineDashDotSetting')
    def dash_size(self) -> Int:
        """
        Get a dash size setting.

        Returns
        -------
        dash_size : Int
            Dash size setting.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_dash_dot_setting = ap.LineDashDotSetting(
        ...     dot_size=2, dash_size=5, space_size=3)
        >>> line.line_dash_dot_setting.dash_size
        Int(5)
        """
        return self['dash_size']

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineDashDotSetting')
    def space_size(self) -> Int:
        """
        Get a space size setting.

        Returns
        -------
        space_size : Int
            Space size setting.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> line: ap.Line = sprite.graphics.draw_line(
        ...     x_start=50, y_start=50, x_end=150, y_end=50)
        >>> line.line_dash_dot_setting = ap.LineDashDotSetting(
        ...     dot_size=2, dash_size=5, space_size=3)
        >>> line.line_dash_dot_setting.space_size
        Int(3)
        """
        return self['space_size']



"""Definitions and interface for loggers.
"""

from logging import INFO
from logging import Formatter
from logging import Logger
from logging import StreamHandler
from logging import getLogger

_LOGGER_NAME_INFO: str = 'user_info'
_is_initial_call: bool = True


def get_info_logger() -> Logger:
    """
    Get information logger.

    Returns
    -------
    logger : Logger
        Information logger as following settings.
        - Level: INFO
        - Format: `%Y-%m-%d %H:%M:%S. <message>`
    """
    global _is_initial_call
    logger: Logger = getLogger(_LOGGER_NAME_INFO)
    logger.setLevel(level=INFO)

    if _is_initial_call:
        stream_handler: StreamHandler = StreamHandler()
        formatter: Formatter = Formatter(
            fmt='%(asctime)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S')
        stream_handler.setFormatter(fmt=formatter)
        logger.addHandler(stream_handler)
        _is_initial_call = False
    return logger



"""Enum class implementation for the easing setting.
"""

from enum import Enum


class Easing(Enum):
    """
    Enum class for the easing setting.

    References
    ----------
    - Easing enum document
        - https://simon-ritchie.github.io/apysc/easing_enum.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=50, height=50)
    >>> _ = rectangle.animation_y(
    ...     y=100,
    ...     duration=1500,
    ...     easing=ap.Easing.EASE_OUT_QUINT,
    ... ).start()
    """
    LINEAR = 'function(x) {return x;}'

    EASE_IN_SINE = 'function(x) {return 1 - Math.cos((x * Math.PI) / 2);}'
    EASE_OUT_SINE = 'function(x) {return Math.sin((x * Math.PI) / 2);}'
    EASE_IN_OUT_SINE = (
        'function(x) {return -(Math.cos(Math.PI * x) - 1) / 2;}'
    )

    EASE_IN_QUAD = 'function(x) {return x * x;}'
    EASE_OUT_QUAD = 'function(x) {return 1 - (1 - x) * (1 - x);}'
    EASE_IN_OUT_QUAD = (
        'function(x) {return x < 0.5 ? 2 * x * x '
        ': 1 - Math.pow(-2 * x + 2, 2) / 2;}'
    )

    EASE_IN_CUBIC = 'function(x) {return x * x * x;}'
    EASE_OUT_CUBIC = 'function(x) {return 1 - Math.pow(1 - x, 3);}'
    EASE_IN_OUT_CUBIC = (
        'function(x) {return x < 0.5 ? 4 * x * x * x '
        ': 1 - Math.pow(-2 * x + 2, 3) / 2;}'
    )

    EASE_IN_QUART = 'function(x) {return x * x * x * x;}'
    EASE_OUT_QUART = 'function(x) {return 1 - Math.pow(1 - x, 4);}'
    EASE_IN_OUT_QUART = (
        'function(x) {return x < 0.5 ? 8 * x * x * x * x '
        ': 1 - Math.pow(-2 * x + 2, 4) / 2;}'
    )

    EASE_IN_QUINT = 'function(x) {return x * x * x * x * x;}'
    EASE_OUT_QUINT = 'function(x) {return 1 - Math.pow(1 - x, 5);}'
    EASE_IN_OUT_QUINT = (
        'function(x) {return x < 0.5 ? 16 * x * x * x * x * x '
        ': 1 - Math.pow(-2 * x + 2, 5) / 2;}'
    )

    EASE_IN_EXPO = (
        'function(x) {return x === 0 ? 0 '
        ': Math.pow(2, 10 * x - 10);}'
    )
    EASE_OUT_EXPO = (
        'function(x) {return x === 1 ? 1 '
        ': 1 - Math.pow(2, -10 * x);}'
    )
    EASE_IN_OUT_EXPO = (
        'function(x) {return x === 0 ? 0 '
        ': x === 1 ? 1 '
        ': x < 0.5 ? Math.pow(2, 20 * x - 10) / 2 '
        ': (2 - Math.pow(2, -20 * x + 10)) / 2;}'
    )

    EASE_IN_CIRC = 'function(x) {return 1 - Math.sqrt(1 - Math.pow(x, 2));}'
    EASE_OUT_CIRC = 'function(x) {return Math.sqrt(1 - Math.pow(x - 1, 2));}'
    EASE_IN_OUT_CIRC = (
        'function(x) {return x < 0.5 '
        '? (1 - Math.sqrt(1 - Math.pow(2 * x, 2))) / 2 '
        ': (Math.sqrt(1 - Math.pow(-2 * x + 2, 2)) + 1) / 2;}'
    )

    EASE_IN_BACK = (
        'function(x) {return 2.70158 * x * x * x - 1.70158 * x * x;}'
    )
    EASE_OUT_BACK = (
        'function(x) {return 1 + 2.70158 * Math.pow(x - 1, 3) '
        '+ 1.70158 * Math.pow(x - 1, 2);}'
    )
    EASE_IN_OUT_BACK = (
        'function(x) {return x < 0.5 '
        '? (Math.pow(2 * x, 2) * ((2.5949095 + 1) * 2 * x - 2.5949095)) / 2 '
        ': (Math.pow(2 * x - 2, 2) * ((2.5949095 + 1) '
        '* (x * 2 - 2) + 2.5949095) + 2) / 2;}'
    )

    EASE_IN_ELASTIC = (
        'function(x) {return x === 0 ? 0 '
        ': x === 1 ? 1 '
        ': -Math.pow(2, 10 * x - 10) * Math.sin((x * 10 - 10.75) '
        '* ((2 * Math.PI) / 3));}')
    EASE_OUT_ELASTIC = (
        'function(x) {return x === 0 ? 0 '
        ': x === 1 ? 1 '
        ': Math.pow(2, -10 * x) * Math.sin((x * 10 - 0.75) '
        '* ((2 * Math.PI) / 3)) + 1;}'
    )
    EASE_IN_OUT_ELASTIC = (
        'function(x) {return x === 0 ? 0 '
        ': x === 1 ? 1 '
        ': x < 0.5 ? -(Math.pow(2, 20 * x - 10) '
        '* Math.sin((20 * x - 11.125) * ((2 * Math.PI) / 4.5))) / 2 '
        ': (Math.pow(2, -20 * x + 10) * Math.sin((20 * x - 11.125) '
        '* ((2 * Math.PI) / 4.5))) / 2 + 1;}'
    )

    EASE_IN_BOUNCE = (
        'function(x) {function ease_out_bounce(y) {'
        'if (y < 1 / 2.75) {return 7.5625 * y * y;} '
        'else if (y < 2 / 2.75) {'
        'return 7.5625 * (y -= 1.5 / 2.75) * y + 0.75;} '
        'else if (y < 2.5 / 2.75) {'
        'return 7.5625 * (y -= 2.25 / 2.75) * y + 0.9375;} '
        'else {return 7.5625 * (y -= 2.625 / 2.75) * y + 0.984375;}} '
        'return 1 - ease_out_bounce(1 - x);}'
    )
    EASE_OUT_BOUNCE = (
        'function(x) {if (x < 1 / 2.75) {return 7.5625 * x * x;} '
        'else if (x < 2 / 2.75) {'
        'return 7.5625 * (x -= 1.5 / 2.75) * x + 0.75;} '
        'else if (x < 2.5 / 2.75) {'
        'return 7.5625 * (x -= 2.25 / 2.75) * x + 0.9375;} '
        'else {return 7.5625 * (x -= 2.625 / 2.75) * x + 0.984375;}}'
    )
    EASE_IN_OUT_BOUNCE = (
        'function(x) {function ease_out_bounce(y) {'
        'if (y < 1 / 2.75) {return 7.5625 * y * y;} '
        'else if (y < 2 / 2.75) {'
        'return 7.5625 * (y -= 1.5 / 2.75) * y + 0.75;} '
        'else if (y < 2.5 / 2.75) {'
        'return 7.5625 * (y -= 2.25 / 2.75) * y + 0.9375;} '
        'else {return 7.5625 * (y -= 2.625 / 2.75) * y + 0.984375;}} '
        'return x < 0.5 ? (1 - ease_out_bounce(1 - 2 * x)) / 2 '
        ': (1 + ease_out_bounce(2 * x - 1)) / 2;}'
    )



"""This module is for the translation mapping data of the
following document:

Document file: display_object_and_graphics_base_prop_abstract.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# DisplayObject and GraphicsBase classes base properties abstract':
    '',

    'This page explains the `DisplayObject` and `GraphicsBase` classes\' each property (such as the x, visible) abstract.':  # noqa
    '',

    '## What apysc can do in its properties':
    '',

    '- You can get/set each property value, such as the x, y, visible.':
    '',

    '## x and y properties':
    '',

    'The x and y properties can get/set the x and y coordinates.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n    direction: ap.Int\n\n\ndef on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : RectOptions\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    direction: ap.Int = options[\'direction\']\n    rectangle.x += direction\n    rectangle.y += direction\n\n    with ap.If(rectangle.x >= 100):\n        direction.value = -1\n        ap.Return()\n\n    with ap.If(rectangle.x <= 50):\n        direction.value = 1\n        ap.Return()\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=200,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\ndirection: ap.Int = ap.Int(1)\noptions: RectOptions = {\'rectangle\': rectangle, \'direction\': direction}\nap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_x_and_y/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_x_and_y/index.html" width="200" height="200"></iframe>\n\nFor more details, please see [DisplayObject class x and y interfaces](display_object_x_and_y.md).':  # noqa
    '',

    '## visible property':
    '',

    'The `visible` property can get/set the visibility of an object.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : RectOptions\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.visible = rectangle.visible.not_\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\noptions: RectOptions = {\'rectangle\': rectangle}\nap.Timer(on_timer, delay=1000, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_visible/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_visible/index.html" width="150" height="150"></iframe>\n\nFor more details, please see [DisplayObject class visible interface](display_object_visible.md).':  # noqa
    '',

    '## rotation interfaces':
    '',

    'The `rotation_around_center` property, `get_rotation_around_point` method, and `set_rotation_around_point` method can get/set the rotation angle.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : RectOptions\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.rotation_around_center += 1\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\noptions: RectOptions = {\'rectangle\': rectangle}\nap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_rotation/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_rotation/index.html" width="150" height="150"></iframe>\n\nFor more details, please see [GraphicsBase class rotation around center interface](graphics_base_rotation_around_center.md) and [GraphicsBase class rotation around point interfaces](graphics_base_rotation_around_point.md).':  # noqa
    '',

    '## scale interfaces':
    '',

    'The `scale_x_from_center` property, `scale_y_from_center` property, `get_scale_x_from_point` method, `set_scale_x_from_point` method, `get_scale_y_from_point` method, and `set_scale_y_from_point` method can get/set the scale values.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n    scale_value: ap.Number\n\n\ndef on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : RectOptions\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    scale_value: ap.Number = options[\'scale_value\']\n    rectangle.scale_x_from_center += scale_value\n    rectangle.scale_y_from_center += scale_value\n\n    with ap.If(rectangle.scale_x_from_center >= 2.0):\n        scale_value.value = -0.01\n        ap.Return()\n\n    with ap.If(rectangle.scale_y_from_center <= 0.5):\n        scale_value.value = 0.01\n        ap.Return()\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nscale_value: ap.Number = ap.Number(0.01)\noptions: RectOptions = {\'rectangle\': rectangle, \'scale_value\': scale_value}\nap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_scale/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_scale/index.html" width="150" height="150"></iframe>\n\nFor more details, please see [GraphicsBase class scale from center interfaces](graphics_base_scale_from_center.md) and [GraphicsBase class scale from point interfaces](graphics_base_scale_from_point.md).':  # noqa
    '',

    '## flip properties':
    '',

    'The `flip_x` and `flip_y` properties can get/set the flip (reflection) setting.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass LineOptions(TypedDict):\n    line: ap.Line\n\n\ndef on_timer(e: ap.TimerEvent, options: LineOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : LineOptions\n        Optional arguments dictionary.\n    """\n    line: ap.Line = options[\'line\']\n    line.flip_x = line.flip_x.not_\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color=\'#fff\', thickness=5)\nline: ap.Line = sprite.graphics.draw_line(\n    x_start=50, y_start=50, x_end=100, y_end=100)\n\noptions: LineOptions = {\'line\': line}\nap.Timer(on_timer, delay=1000, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_flip/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_flip/index.html" width="150" height="150"></iframe>\n\nFor more details, please see [GraphicsBase class flip x and flip y interfaces](graphics_base_flip_interfaces.md).':  # noqa
    '',

    '## skew properties':
    '',

    'The `skew_x` and `skew_y` properties can get/set the skew-value.\n\n<details>\n<summary>Display the code block:</summary>':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : RectOptions\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.skew_x += 1\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\noptions: RectOptions = {\'rectangle\': rectangle}\nap.Timer(on_timer, delay=ap.FPS.FPS_60, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'do_and_graphics_base_prop_abstract_skew/\')\n```':  # noqa
    '',

    '</details>\n\n<iframe src="static/do_and_graphics_base_prop_abstract_skew/index.html" width="150" height="150"></iframe>\n\nFor more details, please see [GraphicsBase class skew x and skew y interfaces](graphics_base_skew.md).':  # noqa
    '',

    '## See also':
    '',

    '- [DisplayObject class](display_object.md)\n- [DisplayObject class parent interfaces](display_object_parent.md)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: boolean.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Boolean':
    '',

    'This page explains the `Boolean` class.\n\nBefore reading on, maybe it is helpful to read the following page:\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the Boolean class?':
    '',

    'The `Boolean` class is the apysc boolean class. It can accept `bool` or `Boolean` values at the constructor, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nbool_1: ap.Boolean = ap.Boolean(True)\nassert bool_1\n\nbool_2: ap.Boolean = ap.Boolean(False)\nassert not bool_2\n\nbool_3: ap.Boolean = ap.Boolean(bool_1)\nassert bool_3\n```':  # noqa
    '',

    '':
    '',

    '## Note for the Bool class alias':
    '',

    'The `Bool` class is the alias of the `Boolean` class. And it behaves the same as the `Boolean` class.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nassert ap.Boolean == ap.Bool\nassert ap.Boolean(True) == ap.Bool(True)\n```':  # noqa
    '',

    '':
    '',

    '## Boolean comparison':
    '',

    'The `Boolean` comparison interface behaves like the Python built-in `bool` value.\n\nYou can compare it with the equal comparison operator (`=`), and the `Boolean`\\, `bool`\\, `int`\\, `Int` types are acceptable, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nbool_1: ap.Boolean = ap.Boolean(True)\nassert bool_1 == True  # noqa\nassert bool_1 == ap.Boolean(True)\nassert bool_1 == 1\nassert bool_1 == ap.Int(1)\n```':  # noqa
    '',

    'Also, the not equal comparison operator (`!=`) is supported, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nbool_1: ap.Boolean = ap.Boolean(True)\nassert bool_1 != False  # noqa\nassert bool_1 != ap.Boolean(False)\nassert bool_1 != 0\nassert bool_1 != ap.Int(0)\n```':  # noqa
    '',

    'You can skip the comparison operator, as follows:':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nbool_1: ap.Boolean = ap.Boolean(True)\nassert bool_1\n\nbool_2: ap.Boolean = ap.Boolean(False)\nassert not bool_2\n```':  # noqa
    '',

    '':
    '',

    '## Reverse a Boolean value':
    '',

    'The `not_` property returns the reversed `Boolean` value:':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nbool_1: ap.Boolean = ap.Boolean(True)\nbool_2: ap.Boolean = bool_1.not_\nassert not bool_2\n\nbool_3: ap.Boolean = bool_2.not_\nassert bool_3\n```':  # noqa
    '',

    '':
    '',

    '## Boolean class constructor API':
    '',

    '<!-- Docstring: apysc._type.boolean.Boolean.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, value:Union[int, apysc._type.int.Int, _ForwardRef(\'Boolean\')]) -> None`<hr>\n\n**[Interface summary]** Boolean class for apysc library.<hr>\n\n**[Parameters]**\n\n- `value`: Boolean or Int or bool or int\n  - Initial boolean value. 0 or 1 are acceptable for an integer value.\n\n<hr>\n\n**[Notes]**\n\nThe Bool class is the alias of the Boolean, and it behaves the same as the Boolean class.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> bool_val_1: ap.Boolean = ap.Boolean(True)\n>>> bool_val_1\nBoolean(True)\n\n>>> bool_val_2: ap.Bool = ap.Bool(True)\n>>> bool_val_2\nBoolean(True)\n```':  # noqa
    '',

    '':
    '',

    '## value property API':
    '',

    '<!-- Docstring: apysc._type.boolean.Boolean.value -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a current boolean value.<hr>\n\n**[Returns]**\n\n- `value`: bool\n  - Current boolean value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> bool_val: ap.Boolean = ap.Boolean(True)\n>>> bool_val.value = False\n>>> bool_val.value\nFalse\n\n>>> bool_val.value = ap.Boolean(True)\n>>> bool_val.value\nTrue\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [apysc fundamental data classes value interface](https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html)':  # noqa
    '',

    '## not_ property API':
    '',

    '<!-- Docstring: apysc._type.boolean.Boolean.not_ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a not condition Boolean value.<hr>\n\n**[Returns]**\n\n- `result`: Boolean\n  - Not condition Boolean value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> bool_val: ap.Boolean = ap.Boolean(True)\n>>> bool_val.not_\nBoolean(False)\n\n>>> bool_val.value = False\n>>> bool_val.not_\nBoolean(True)\n```':  # noqa
    '',

}



"""Class implementation for the blank object interface.
The apysc uses this interface at the custom event interface
or something else.
"""

from apysc._html.debug_mode import add_debug_info_setting


class BlankObjectInterface:

    _is_blank_object_initialized: bool = False
    _blank_object_variable_name: str

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='BlankObjectInterface')
    def _initialize_blank_object_if_not_initialized(self) -> None:
        """
        Initialize a blank object value if this interface
        does not initialize it yet.
        """
        from apysc._expression import expression_data_util
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._expression.event_handler_scope import \
            TemporaryNotHandlerScope
        if self._is_blank_object_initialized:
            return
        with TemporaryNotHandlerScope():
            self._blank_object_variable_name = expression_variables_util.\
                get_next_variable_name(type_name=var_names.BLANK_OBJECT)
            expression: str = (
                f'var {self._blank_object_variable_name} = {{}};'
            )
            expression_data_util.append_js_expression(
                expression=expression)
            self._is_blank_object_initialized = True

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='BlankObjectInterface')
    def blank_object_variable_name(self) -> str:
        """
        Get a blank object variable name.

        Returns
        -------
        blank_object_variable_name : str
            A blank object variable name.
        """
        self._initialize_blank_object_if_not_initialized()
        return self._blank_object_variable_name



"""This module is for the translation mapping data of the
following document:

Document file: elif.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Elif':
    '',

    'This page explains the `Elif` class.\n\nBefore reading on, maybe it is helpful to read the following page (the apysc uses the `Elif` class for the same reason of each other data type):\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the Elif class?':
    '',

    'The `Elif` class is the apysc branch instruction class. It behaves like the Python built-in `elif` keyword.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `Elif` class requires using the `with` statement. Also, The `Elif` class statement is only acceptable to implement right after the `If` or `Elif` classes statement.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\n\ncondition_1: ap.Boolean = ap.Boolean(False)\ncondition_2: ap.Boolean = ap.Boolean(True)\nwith ap.If(condition_1):\n    int_1 += 20\nwith ap.Elif(condition_2):\n    int_1 += 30\n```':  # noqa
    '',

    '':
    '',

    '## Notes':
    '',

    'If you insert the code between the `If` (or `Elif`) and `Elif` statements, then it raises exceptions:':  # noqa
    '',

    '```py\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\n\ncondition_1: ap.Boolean = ap.Boolean(False)\ncondition_2: ap.Boolean = ap.Boolean(True)\nwith ap.If(condition_1):\n    int_1 += 20\n# Code inserting between the `If` and `Elif` will raise an exception.\nint_2: ap.Int = ap.Int(30)\nwith ap.Elif(condition_2):\n    int_1 += 30\n```':  # noqa
    '',

    '':
    '',

    '```\nValueError: Elif interface can only use right after If or Elif interfaces.\n```':  # noqa
    '',

    'Also, you can\'t create the condition (`Boolean` value) at the `Elif` constructor position (the same goes for the comparison operators), for instance:':  # noqa
    '',

    '```py\nimport apysc as ap\n\nint_1: ap.Int = ap.Int(10)\n\ncondition_1: ap.Boolean = ap.Boolean(False)\ncondition_2: ap.Boolean = ap.Boolean(True)\nwith ap.If(condition_1):\n    int_1 += 20\nwith ap.Elif(int_1 == 10):\n    int_1 += 30\n```':  # noqa
    '',

    '':
    '',

    '```\nValueError: Elif interface can only use right after If or Elif interfaces.\n\nMaybe you are using Int or String, or anything else comparison expression at Elif constructor (e.g., `with Elif(any_value == 10, ...):`).\nCurrently, that specifying expression directly is not supported so please define condition separately as follows:\ncondition: Boolean = any_value == 10\n...\nwith Elif(condition, ....):\n```':  # noqa
    '',

    '':
    '',

    '## See also':
    '',

    '- [If class](if.md)\n- [Else class](else.md)\n- [Each branch instruction class\'s scope variables reverting setting](branch_instruction_variables_reverting_setting.md)':  # noqa
    '',

    '## Elif constructor API':
    '',

    '<!-- Docstring: apysc._branch._elif.Elif.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, condition:Union[apysc._type.boolean.Boolean, NoneType], *, locals_:Union[Dict[str, Any], NoneType]=None, globals_:Union[Dict[str, Any], NoneType]=None) -> None`<hr>\n\n**[Interface summary]** A class to append the `else if` branch instruction expression.<hr>\n\n**[Parameters]**\n\n- `condition`: Boolean or None\n  - Boolean value to be used for judgment.\n- `locals_`: dict or None, default None\n  - Current scope\'s local variables. Set locals() value to this argument. If specified, this interface reverts all local scope VariableNameInterface variables (like Int, Sprite) at the end of an `Elif` scope. This setting is useful when you don\'t want to update each variable by implementing the `Elif` scope.\n- `globals_`: dict or None, default None\n  - Current scope\'s global variables. Set globals() value to this argument. This setting works the same way as the locals_ argument.\n\n<hr>\n\n**[Notes]**\n\n ・Currently the apysc can not initialize condition value in the constructor. For example, the following raises exception: <br> ・You can only use this class immediately after the `If` or `Elif` statement. ```py import apysc as ap with ap.Elif(any_value >= 10): ... ``` You can avoid this exception by predefining condition value, as follows: >>> import apysc as ap >>> any_value: ap.Int = ap.Int(10) >>> condition_1: ap.Boolean = any_value >= 10 >>> condition_2: ap.Boolean = any_value >= 5 >>> with ap.If(condition_1): ... # Do something here ... pass >>> with ap.Elif(condition_2): ... # Do something else here ... pass<hr>\n\n**[References]**\n\n- [Each branch instruction class\'s scope variables reverting setting](https://simon-ritchie.github.io/apysc/branch_instruction_variables_reverting_setting.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_return_value.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Each animation interface return value':
    '',

    'This page explains each animation interface, such as the `animation_move`\\, return value (`AnimationBase` instance).':  # noqa
    '',

    '## Each interface returns the subclass instance of the AnimationBase':
    '',

    'Each animation interface returns the subclass instance of the `AnimationBase`\\. So, for example, the `animation_move` interface returns the `AnimationMove` instance, and the `animation_x` interface  returns the `AnimationX` instance.\n\nThe `AnimationBase` class has the standard animation interfaces, such as the `start` (method to start animation), `animation_complete` (method to bind the animation completion event), `target` (property of the animation target).':  # noqa
    '',

    '## Basic usage':
    '',

    'Each return value class is in the apysc package (e.g., `ap.AnimationMove`). Therefore, you can set the type annotation with it.\n\nThe following code example uses the `animation_x` method interface. And You get an `AnimationX` instance to start and bind the animation completion event with it.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nDURATION: int = 1000\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=50, duration=DURATION)\n    animation_x.animation_complete(on_animation_complete_2)\n    animation_x.start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle],\n        options: dict) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=100, duration=DURATION)\n    animation_x.animation_complete(on_animation_complete_1)\n    animation_x.start()\n\n\nap.Stage(\n    stage_width=200, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nanimation_x: ap.AnimationX = rectangle.animation_x(\n    x=100, duration=DURATION)\nanimation_x.animation_complete(on_animation_complete_1)\nanimation_x.start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_return_value_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_return_value_basic_usage/index.html" width="200" height="150"></iframe>':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_base_rotation_around_point.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# GraphicsBase rotation_around_point interfaces':
    '',

    'This page explains the `GraphicsBase` class (base class of each graphics, such as the `Rectangle`) `get_rotation_around_point` and `set_rotation_around_point` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `get_rotation_around_point` method will return a rotation value around the given coordinates, and the `set_rotation_around_point` method will update a rotation value around the given coordinates.\n\nThese rotation values are relative, and each point has the rotation value. For example, the coordinates of the `(x=50, y=50)` and the other coordinates of the `(x=100, y=100)` have different rotation values.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `get_rotation_around_point` method requires the `x` and `y` arguments and return a rotation value around the given coordinates. The `set_rotation_around_point` requires `x`, `y` and `rotation` arguments. All the arguments and return value are `Int` type.\n\nThe following example creates the two rectangles and rotates each rectangle in the timer handler. The first rectangle rotates around the top-left coordinates (`x=50, y=50`). Also, the second one rotates around the bottom-right coordinates (`x=100, y=100`).':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectanglesOptions(TypedDict):\n    rectangle_1: ap.Rectangle\n    rectangle_2: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: _RectanglesOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    x: ap.Int = ap.Int(50)\n    y: ap.Int = ap.Int(50)\n    rectangle_1: ap.Rectangle = options[\'rectangle_1\']\n    rotation: ap.Int = rectangle_1.get_rotation_around_point(x=x, y=y)\n    rotation += 1\n    rectangle_1.set_rotation_around_point(rotation=rotation, x=x, y=y)\n\n    rectangle_2: ap.Rectangle = options[\'rectangle_2\']\n    x = ap.Int(100)\n    y = ap.Int(100)\n    rotation = rectangle_2.get_rotation_around_point(x=x, y=y)\n    rotation += 1\n    rectangle_2.set_rotation_around_point(rotation=rotation, x=x, y=y)\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\n\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\noptions: _RectanglesOptions = {\n    \'rectangle_1\': rectangle_1, \'rectangle_2\': rectangle_2}\ntimer: ap.Timer = ap.Timer(\n    on_timer, delay=ap.FPS.FPS_60, options=options)\ntimer.start()\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_base_rotation_around_point_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_base_rotation_around_point_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## get_rotation_around_point API':
    '',

    '<!-- Docstring: apysc._display.rotation_around_point_interface.RotationAroundPointInterface.get_rotation_around_point -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `get_rotation_around_point(self, x:apysc._type.int.Int, y:apysc._type.int.Int) -> apysc._type.int.Int`<hr>\n\n**[Interface summary]** Get a rotation value around the given coordinates.<hr>\n\n**[Parameters]**\n\n- `x`: Int\n  - X-coordinate.\n- `y`: Int\n  - Y-coordinate.\n\n<hr>\n\n**[Returns]**\n\n- `rotation`: Int\n  - Rotation value around the given coordinates.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> x: ap.Int = ap.Int(100)\n>>> y: ap.Int = ap.Int(100)\n>>> rectangle.set_rotation_around_point(\n...     rotation=ap.Int(45), x=x, y=y)\n>>> rectangle.get_rotation_around_point(x=x, y=y)\nInt(45)\n```':  # noqa
    '',

    '':
    '',

    '## set_rotation_around_point API':
    '',

    '<!-- Docstring: apysc._display.rotation_around_point_interface.RotationAroundPointInterface.set_rotation_around_point -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `set_rotation_around_point(self, rotation:apysc._type.int.Int, x:apysc._type.int.Int, y:apysc._type.int.Int) -> None`<hr>\n\n**[Interface summary]** Update a rotation value around the given coordinates.<hr>\n\n**[Parameters]**\n\n- `rotation`: Int\n  - Rotation value to set.\n- `x`: Int\n  - X-coordinate.\n- `y`: Int\n  - Y-coordinate.':  # noqa
    '',

}



"""Class implementation for the skew x interface.
"""

from typing import Dict

from apysc._animation.animation_skew_x_interface import AnimationSkewXInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class SkewXInterface(
        AnimationSkewXInterface, RevertInterface, AttrLinkingInterface):

    _skew_x: Int

    def _initialize_skew_x_if_not_initialized(self) -> None:
        """
        Initialize the _skew_x attribute if this instance does not
        initialize it yet.
        """
        if hasattr(self, '_skew_x'):
            return
        self._skew_x = Int(0)

        self._append_skew_x_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='SkewXInterface')
    def _append_skew_x_attr_linking_setting(self) -> None:
        """
        Append a skew-x attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._skew_x, attr_name='skew_x')
        self._append_attr_to_linking_stack(
            attr=self._skew_x, attr_name='skew_x')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='SkewXInterface')
    def skew_x(self) -> Int:
        """
        Get a current skew x value of the instance.

        Returns
        -------
        skew_x : Int
            Current skew x value of this instance.

        References
        ----------
        - GraphicsBase skew_x and skew_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_skew.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.skew_x = ap.Int(50)
        >>> rectangle.skew_x
        Int(50)
        """
        from apysc._type import value_util
        self._initialize_skew_x_if_not_initialized()
        return value_util.get_copy(value=self._skew_x)

    @skew_x.setter
    def skew_x(self, value: Int) -> None:
        """
        Update a skew x value of this instance.

        Parameters
        ----------
        value : Int
            Skew x value to set.

        References
        ----------
        - GraphicsBase skew_x and skew_y interfaces document
            - https://simon-ritchie.github.io/apysc/graphics_base_skew.html
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='skew_x', args=[value], kwargs={},
                module_name=__name__,
                class_name=SkewXInterface.__name__):
            import apysc as ap
            from apysc._validation import number_validation
            self._initialize_skew_x_if_not_initialized()
            number_validation.validate_integer(integer=value)
            before_value: ap.Int = self._skew_x
            self._skew_x = value
            self._append_skew_x_update_expression(before_value=before_value)

            self._append_skew_x_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='SkewXInterface')
    def _append_skew_x_update_expression(
            self, *, before_value: Int) -> None:
        """
        Append the skew x updating expression.

        Parameters
        ----------
        before_value : ap.Int
            Before updating value.
        """
        import apysc as ap
        from apysc._type import value_util
        before_value_str: str = value_util.get_value_str_for_expression(
            value=before_value)
        after_value_str: str = value_util.get_value_str_for_expression(
            value=self._skew_x)
        expression: str = (
            f'{self.variable_name}.skew(-{before_value_str}, 0);'
            f'\n{self.variable_name}.skew({after_value_str}, 0);'
            f'\n{before_value_str} = {after_value_str};'
        )
        ap.append_js_expression(expression=expression)

    _skew_x_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_skew_x_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_skew_x_snapshots',
            value=int(self._skew_x._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._skew_x._value = self._skew_x_snapshots[snapshot_name]
